import Pyro.core
import Pyro.naming
from Pyro.errors import *
import Pyro.util
import time, types, re, os

EVENTSERVICE_NAME = ":Pyro.EventService"
EVENTSERVICE_VERSION = "2.4"


if not Pyro.util.supports_multithreading():
	raise NotImplementedError('Event Service needs Python threads support.')


from threading import Semaphore, Thread

# a lock class for thread synchronization.
class Lock:
	def __init__(self):
		self.sema=Semaphore()
	def lock(self):
		self.sema.acquire()
	def release(self):
		self.sema.release()



# EVENT - the thing that is published. Has a subject and contains a message.
class Event:
	def __init__(self, subject, msg, creationTime=None):
		self.msg=msg
		self.subject=subject
		self.time=creationTime or time.time()
	def __str__(self):
		return "<EVENT SUBJ %s (%s): %s>" % (self.subject, time.ctime(self.time), str(self.msg))


# The EVENTSERVICE is the actual Pyro server.
# It processes the requests in separate threads,
# synchronized by a lock. Otherwise it would be
# possible to call the ES again while receiving a 
# event, thus causing troubles or socket deadlock.
#
# BTW: Subscribers are remembered trough their proxy class.
# This class is capable of being a correct key in a dictionary.
class EventService(Pyro.core.ObjBase):
	def __init__(self):
		Pyro.core.ObjBase.__init__(self)
		self.subscribers={}			# subject -> { subscriber -> dummy }
		self.subscribersMatch={}	# subjectPattern -> { subscriber -> dummy }
		self.lock=Lock()
	def _mksequence(self, seq):
		if not (type(seq) in (types.TupleType,types.ListType)):
			seq=(seq,)
		return seq
	def subscribe(self, subjects, subscriber):
		if subjects: Thread(target=self._do_subscribe, args=(subjects,subscriber)).start()
	def subscribeMatch(self, subjects, subscriber):
		if subjects: Thread(target=self._do_subscribeMatch, args=(subjects,subscriber)).start()
	def unsubscribe(self, subjects, subscriber):
		if subjects: Thread(target=self._do_unsubscribe, args=(subjects,subscriber)).start()
	def publish(self, subjects, message):
		if subjects: Thread(target=self._do_publish, args=(subjects,message)).start()
	def _do_subscribe(self, subjects, subscriber):
		self.lock.lock()
		subscriber._setOneway('event') # don't wait for answer for this method
		# Subscribe into a dictionary; this way; somebody can subscribe
		# only once to this subject. Subjects are exact strings.
		for subject in self._mksequence(subjects):
			self.subscribers.setdefault(subject.lower(),{}) [subscriber]=1
		self.lock.release()	
	def _do_subscribeMatch(self, subjects, subscriber):
		self.lock.lock()
		subscriber._setOneway('event') # don't wait for answer for this method
		# Subscribe into a dictionary; this way; somebody can subscribe
		# only once to this subject. Subjects are regex patterns.
		for subject in self._mksequence(subjects):
			matcher = re.compile(subject,re.IGNORECASE)
			self.subscribersMatch.setdefault(matcher,{}) [subscriber]=1
		self.lock.release()	
	def _do_unsubscribe(self, subjects, subscriber):
		self.lock.lock()
		for subject in self._mksequence(subjects):
			try:
				del self.subscribers[subject.lower()] [subscriber]
			except KeyError:
				try:
					m=re.compile(subject,re.IGNORECASE)
					del self.subscribersMatch[m] [subscriber]
				except KeyError:
					pass
		self.lock.release()			
	def _do_publish(self, subjects, message):
		# keep the creation time, this must be the same for all events.
		creationTime=time.time()
		# now enter the lock.
		self.lock.lock()
		# publish a message. Subjects must be exact strings
		for subject in self._mksequence(subjects):
			event = Event(subject, message, creationTime)
			subjectLC=subject.lower()
			try:
				for s in self.subscribers[subjectLC].keys():
					try:
						s.event(event) # this is a oneway invocation, no result
					except ConnectionClosedError:
						# hmm, subscriber no longer connected...
						del self.subscribers[subjectLC] [s]
					except:
						# ignore all other errors
						pass
			except KeyError:
				pass
			# process the subject patterns
			for (m,subs) in self.subscribersMatch.items():
				if m.match(subject):
					# send event to all subscribers
					for s in subs.keys():	
						try:
							s.event(event)
						except ConnectionClosedError:
							# hmm, subscriber no longer connected...
							del self.subscribersMatch[m] [s]
						except:
							# ignore all other errors
							pass
		self.lock.release()			


def start():
	print '*** Pyro Event Service ***'
	Pyro.core.initServer()
	PyroDaemon = Pyro.core.Daemon()
	locator = Pyro.naming.NameServerLocator()
	print 'Looking for Name Server...'
	ns = locator.getNS(host=Pyro.config.PYRO_NS_HOSTNAME)
	PyroDaemon.useNameServer(ns)

	try:
		ns.unregister(EVENTSERVICE_NAME)
	except NamingError:
		pass

	PyroDaemon.connect(EventService(), EVENTSERVICE_NAME)
	print 'Event service V'+EVENTSERVICE_VERSION,'is running.'
	while 1:
		PyroDaemon.handleRequests()

